#pragma once
/**
 * \brief	This file implement the functionality offered by the chi2testtable function of the r code.
 */

//Includeing Yggdrasil core
#include <core/yggdrasil_core.hpp>
#include <core/yggdrasil_assert.hpp>
#include <core/yggdrasil_exception.hpp>
#include <core/yggdrasil_int.hpp>
#include <core/yggdrasil_valueArray.hpp>

#include <util/yggdrasil_intervalBound.hpp>
#include <util/yggdrasil_hyperCube.hpp>
#include <util/yggdrasil_generalBinEdges.hpp>

#include <interfaces/yggdrasil_interface_parametricModel.hpp>

#include <samples/yggdrasil_dimensionArray.hpp>
#include <samples/yggdrasil_sample.hpp>



//Incluide std
#include <numeric>
#include <algorithm>

//Include boost
#include <boost/math/distributions/chi_squared.hpp>
#include <boost/math/distributions/normal.hpp>


YGGDRASIL_NS_BEGIN(yggdrasil)


/**
 * \class 	yggdrasil_chi2TestTable_t
 * \brief	This class stores a test table similar tot he one that
 * 		 are generated by chi2testtable function of the R package.
 *
 * This table can then be used to compute the test statistics for the GOF
 * tests. This is an internal class, so it may be ugly.
 *
 * The table does not opperate on empirical quantiles as the R-Code
 * does but on theoretical quantiles.
 *
 * It basically owns an array which counts the observed samples
 * in each bin and also own a binned range.
 *
 * It supports two methods of of binning.
 * One that operates with comparissons, but requiers k * N many
 * operations per dimension, where k is the number of bins and N
 * is the number of samples.
 *
 * There is also a fast one, that only needs O[N * Log(N)] many
 * operation but requiers an array of soize N.
 *
 * FOr that This class uses a unsigned int (32Bit) for indexing.
 * This can be changed if greater samples need to be treaded.
 *
 * This class uses a saturating scheme, that means that the
 * edge boundaries are not strictly enforced and the
 * last and the fist bin are implicitly a bit larger.
 */
class yggdrasil_chi2TestTable_t
{
	/*
	 * ===================
	 * Typedefs
	 */
public:
	using HyperCube_t 	= yggdrasil_hyperCube_t;	//!< This is the type that represents a domain.
	using Numeric_t 	= ::yggdrasil::Numeric_t;	//!< This is the underling numeric type.
	using Real_t 		= ::yggdrasil::Real_t;		//!< This is the high precission type.
	using Size_t 		= ::yggdrasil::Size_t;
	using Sample_t 		= yggdrasil_sample_t;		//!< This is the type of a single sample.
	using DimArray_t 	= yggdrasil_dimensionArray_t;	//!< This is the type of the diomensional array.
	using IntervalBound_t 	= HyperCube_t::IntervalBound_t;	//!< This is the bound for oneinterval.
	using ParModel_t 	= yggdrasil_parametricModel_i;	//!< This is the type ofr the parametric model
	using ParModel_ptr 	= yggdrasil_parametricModel_i*;	//!< This is a pointer to the parameteric model.
	using cParModel_ptr 	= const yggdrasil_parametricModel_i*;	//!< This is the constant version of the model pointer.
	using ValueArray_t 	= ParModel_t::ValueArray_t;	//!> This is the type for storing the expected count

	using SortIndex_t 	= ::yggdrasil::SortIndex_t;


	/*
	 * The typedefes are derived from the edges
	 */
	using BinEdgesArray_t   = yggdrasil_genBinEdges_t;	//!< This is the class that is used for the bin edges
	using BinEdge_it 	= BinEdgesArray_t::const_iterator;	//!> This is the type that allows iterating over the bin edges
	using BinCountArray_t 	= yggdrasil_valueArray_t<Size_t>;//!< This is the array for counting the occurances of the samples is a bin.
	using BinCount_it 	= BinCountArray_t::const_iterator;	//!> This is an iterator to access the counts in a bin

	/*
	 * =================
	 * Constructors
	 *
	 * All constructers are defaulted.
	 */
public:
	/**
	 * \brief	Building constructor.
	 *
	 * Thsi constructor is used to construct an empty,
	 * not associated table for a chi2 test.
	 *
	 * The user has to supply the significance level,
	 * that is scalled in the range 0 to 1.
	 * The user also has to specify if the table is
	 * used for multi-variate samples (meaning inside
	 * the independence test) or not.
	 * This value will determine the choice of the
	 * numbers of bins.
	 * The actual number of bins depends on the
	 * data that is, in a later state associated to
	 * *this.
	 *
	 * It is important that the data is not yet bind
	 * *this.
	 *
	 * \param  sigLevel	  	The significance niveau that should be used.
	 * \param  isMultiVariate	If a different rule for the bins should be used.
	 *
	 * \throw	This constructor checks if the sigLevel is smaller/equal zero.
	 * 		 In that case an exception is geerated. also if the level is greater/equal one.
	 */
	yggdrasil_chi2TestTable_t(
		const Numeric_t 	sigLevel,
		const bool 		isMultiVariate)
	 :
	  m_binCounts(),
	  m_bins(),
	  m_sigLevel(sigLevel),
	  m_isMultiVariate(isMultiVariate)
	{
		if(m_sigLevel <= 0.0)
		{
			throw YGGDRASIL_EXCEPT_InvArg("The significance level is to small, it was " + std::to_string(m_sigLevel));
		};
		if(1.0 <= m_sigLevel)
		{
			throw YGGDRASIL_EXCEPT_InvArg("The significance level is to large, it was " + std::to_string(m_sigLevel));
		};
	}; //End: building constructor


	/**
	 * \brief	This constructor is a safe guiard
	 */
	yggdrasil_chi2TestTable_t(
		const bool 	 	isMultiVariate,
		const Numeric_t 	sigLevel)
	 = delete;


	/**
	 * \brief	Default constructor.
	 *
	 * Is forbidden.
	 */
	yggdrasil_chi2TestTable_t()
	 = delete;

	/**
	 * \brief	Copy constructor.
	 *
	 * Is defaulted.
	 */
	yggdrasil_chi2TestTable_t(
		const yggdrasil_chi2TestTable_t&)
	 = default;

	/**
	 * \brief	Move constructor.
	 *
	 * Is defaulted.
	 */
	yggdrasil_chi2TestTable_t(
		yggdrasil_chi2TestTable_t&&)
	 = default;

	/**
	 * \brief	Copy assignment.
	 *
	 * Is defaulted.
	 */
	yggdrasil_chi2TestTable_t&
	operator= (
		const yggdrasil_chi2TestTable_t&)
	 = default;

	/**
	 * \brief	Move assignment.
	 *
	 * Is defaulted
	 */
	yggdrasil_chi2TestTable_t&
	operator= (
		yggdrasil_chi2TestTable_t&&)
	 = default;


	/**
	 * \brief	Destructor.
	 *
	 * Is defaulted.
	 */
	~yggdrasil_chi2TestTable_t()
	 = default;

	/*
	 * ================
	 * Utility Functions
	 */
public:
	/**
	 * \brief	Return the number of bins \c *this has.
	 *
	 * \throw 	if \c *this is not associated to any data.
	 */
	Size_t
	nBins()
	 const
	{
		if(this->isAssociated() == false)
		{
			throw YGGDRASIL_EXCEPT_LOGIC("A table without associated data has an undefined numbers of bins.");
		};

		yggdrasil_assert(m_binCounts.size() == m_bins.nBins());

		return m_binCounts.size();
	};


	/**
	 * \brief	This function returns true if \c *this is associated to a domain.
	 *
	 * A state is assocciated to data if an initaial fitting was performed on it.
	 * The net effect is that the \e bin \e count, as reported by nBins()
	 * is different form zero.
	 */
	bool
	isAssociated()
	 const
	 noexcept
	{
		yggdrasil_assert((m_binCounts.size() != 0) == m_bins.isAssociated()); //Test if they are in sync
		return ((m_binCounts.empty() == true) && (m_bins.isAssociated() == false)) ? false : true;
	};


	/**
	 * \brief	Returns the significants level \c *this was constructed with.
	 *
	 * It is not requiered that \c *this is associiated to any data.
	 */
	Numeric_t
	getSigLevel()
	 const
	 noexcept
	{
		return m_sigLevel;
	};


	/**
	 * \brief	Returns true if \c *this was created for multi-variate data.
	 *
	 * It is not requiered that \c *this is associiated to any data.
	 */
	bool
	isForMultiVariate()
	 const
	 noexcept
	{
		return m_isMultiVariate;
	};





	/**
	 * \brief	This function returns the number of samples
	 * 		 that are recoreded inside \c *this.
	 *
	 * This function basically sum up all the bin values.
	 *
	 * \throw	If *this is not associated to any data.
	 */
	Size_t
	countSamples()
	 const
	{
		if(this->isAssociated() == false)
		{
			throw YGGDRASIL_EXCEPT_LOGIC("The GOF CHI2 state is not associated ot data.");
		};

		Size_t accum1 = 0,
		       accum2 = 0,
		       accum3 = 0,
		       accum4 = 0;

		const int nBin      = m_binCounts.size();
		const int nBin_mod4 = nBin % 4;
		const int N         = nBin - nBin_mod4;
		for(int i = 0; i != N; i += 4)
		{
			accum1 += m_binCounts[i + 0];
			accum2 += m_binCounts[i + 1];
			accum3 += m_binCounts[i + 2];
			accum4 += m_binCounts[i + 3];
		}; //End for(i): accumulation

		for(int i = N; i != nBin; ++i)
		{
			accum1 += m_binCounts[i];
		}; //End for(i): Handling the left over

		//Final sum up and return

		return ((accum1 + accum2) + (accum3 + accum4));
	}; //End: nSamples



	/*
	 * ===================
	 * Accessing functions
	 */
public:
	/**
	 * \brief	Returns the count of the \c i th bin.
	 *
	 * This function returns the number of samples that are
	 * recorded in the bin \c i .
	 *
	 * \param  i 	The bin we want to access.
	 *
	 * \throw	If i is out of range.
	 */
	Size_t
	getCountInBin(
		const uInt_t 	i)
	 const
	{
		if(this->isAssociated() == false)
		{
			throw YGGDRASIL_EXCEPT_LOGIC("A table without associated data, does not have bins.");
		};

		if(i >= m_binCounts.size())
		{
			throw YGGDRASIL_EXCEPT_OutOfBound("Tried to access the count of bin " + std::to_string(i) + ", but there are only " + std::to_string(m_binCounts.size()) + " many bins");
		};

		return m_binCounts[i];
	}; //End: getCountInBin


	/**
	 * \brief	This function returns the interval that is spanned by the \c i th bin.
	 *
	 * \param  i 	Which interval should be queried
	 *
	 * \throw	If \c *this is not associated with any data.
	 * 		 i is out of range or if the interval that is constructed is invalid.
	 */
	IntervalBound_t
	getBinInterval(
		const uInt_t 	i)
	 const
	{
		//All the tests are performed inside m_bins object
		const IntervalBound_t binInterval = m_bins.getBinInterval(i);

		yggdrasil_assert(binInterval.isValid() == true);

		//Return the interval
		return binInterval;
	}; //End: getBinInterval


	/**
	 * \brief	Returns the iterator to the counting of the first bin.
	 *
	 * Notice that this function only guarantees read only accesss.
	 */
	BinCount_it
	begin()
	 const
	 noexcept
	{
		return m_binCounts.cbegin();
	};


	/**
	 * \brief	Returns the past the end iterator of the count.
	 *
	 * Noticte that this functin also allows read only access.
	 */
	BinCount_it
	end()
	 const
	 noexcept
	{
		return m_binCounts.cend();
	};


	/**
	 * \brief	This fucntion returns a constant reference to the bin edges of \c *this.
	 *
	 * For that the standard bin edge object is used that is provided by Yggdrasil.
	 */
	const BinEdgesArray_t&
	getBinEdges()
	 const
	{
		if(this->isAssociated() == false)
		{
			throw YGGDRASIL_EXCEPT_LOGIC("Can not acccess an table that is not associated.");
		};

		return m_bins;
	}; //End: getBinEdges




	/*
	 * ================
	 * Managing functions
	 *
	 * These functions allows to manage *this.
	 * Like adding samples or so.
	 */
public:

	/**
	 * \brief	This function associated \c *this with
	 * 		 the data in dmensional array \c dimArray.
	 *
	 * This function sets up *this.
	 * It will associate itself to the dimensional array.
	 * It will counts the number of smaples that falls into a specific bin.
	 * It uses the same rools to descide the number of the bins
	 * as the R-Code does. However it uses theoretical
	 * quantiles, that are specified by the parametric model to determine
	 * the bin edges.
	 *
	 * It is possible to use a fast versoion of the determing vewrsion.
	 * But it will occupy a lot of memory.
	 *
	 * \param  d 		This is the dimension which we operate on.
	 * \param  domain	This is the domain, in rescalled root data space, that *this occupies.
	 * \param  dimArray	This is the dimensional array that contains teh sample.
	 * \param  parModel	This is the parametric model, that describes the sample inside the domain.
	 * \param  useSorting	Set to true to use the faster algorithm that requiers possible lot of memeory.
	 *
	 * TODO: What is with the serious problems?
	 *
	 * \throw 	This function throws if \c *thsi si allready associated to data.
	 * 		 And if serious problems where detected.
	 */
	bool
	buildInitialChi2Table(
		const uInt_t 			d,
		const IntervalBound_t&		domain,
		const DimArray_t&		dimArray,
		cParModel_ptr 			parModel,
		const bool 			useSorting);


	/*
	 * \brief	This function associated \c *this with
	 * 		 the data in dmensional array \c dimArray.
	 *
	 * This function is simillar to the one that takes a model,
	 * so please also see there for more information.
	 * This fucntion however is more flexible, since the binning
	 * is passed explicit to *this.
	 *
	 * \param  dimArray	This is the dimensional array that contains teh sample.
	 * \param  binEdges	This is the domain and the binning that should be applied.
	 * \param  useSorting	Set to true to use the faster algorithm that requiers possible lot of memeory.
	 *
	 * TODO: What is with the serious problems?
	 *
	 * \throw 	This function throws if \c *thsi si allready associated to data.
	 * 		 And if serious problems where detected.
	 *
	 */
	bool
	internal_buildInitialChi2TableDirect(
		const DimArray_t&		dimArray,
		const BinEdgesArray_t& 		bins,
		const bool 			useSorting);



	/**
	 * \brief	This function returns a copy of \c *this, but not associated to any data.
	 *
	 * This function calls the building constructor.
	 * The parameters are the same as were used to construct \c *this.
	 *
	 * Thsi function works even if *this is not associated
	 */
	yggdrasil_chi2TestTable_t
	getUnassociatedCopy()
	 const
	{
		return yggdrasil_chi2TestTable_t(m_sigLevel, m_isMultiVariate);
	};



	/**
	 * \brief	This function unbounds *this from the associated data.
	 *
	 * This is basically the same as clearing the internal data.
	 * Or it is like creating an unaccociated copy and assigning it to *this.
	 * Not that this does not clear the significance level or the
	 * Multivar level bool.
	 * It just clears the internal memory.
	 *
	 * \throw	This must be associated to data.
	 */
	void
	unboundData();


	/*
	 * =================
	 * Statistical Functions
	 */
public:
	/**
	 * \brief	This function computes the CHI2 Value of the test statistics.
	 *
	 * This is \f$ \sum_{i} = \frac{(O_i - E_i)^2}{E_i} \f$, where O_i is the number of
	 * observations that falls into class i and E_i is the number of expected observations,
	 * that falls into class i.
	 *
	 * This function also handles some special cases. For example if E_i and O_i both
	 * are zero then we have 0 / 0 which is supposed to return nan.
	 * However in the statistcal sense, it should be zero, since the expectaion and the observation
	 * are compatible with each other.
	 *
	 * This function can also return infinity, if E_i == 0 but O_i != 0, then we have in
	 * finity rejecction. This means that this should not happens.
	 * An implementation note, if that happens the loop (sum) exits earlier and the
	 * other values are not checked.
	 *
	 * \param  expectCount		This is the expected count. Must be positiv or zero and
	 * 				 not infinity or nan.
	 *
	 * \note	If NDEBUG is not defined, then all values are checked. before the iteration.
	 *
	 * \throw	If the imput contains a non compatible value or if NAN or a negative value is generated.
	 */
	Real_t
	computeChi2Value(
		const ValueArray_t& 	expCount)
	 const;




	/*
	 * ================
	 * Private functions
	 */
private:
	/**
	 *
	 * \brief	This fucntion fills the number of bins using a slow algorithm.
	 *
	 * The algorithm may be slow, but it requieres almost no memeory.
	 *
	 * \return 	It returns true, if it was successfull
	 */
	bool
	priv_distributeToBins_std(
		const DimArray_t&	dimArray);

	/**
	 *
	 * \brief	This fucntion fills the number of bins using a sorting algorithm.
	 *
	 * However this algorithm needs some helper storage, a lot of them.
	 *
	 * \return 	It returns true, if it was successfull
	 */
	bool
	priv_distributeToBins_sort(
		const DimArray_t&	dimArray);



	/*
	 * =================
	 * Private Members
	 */
private:
	BinCountArray_t		m_binCounts;	//!< This vector counts the samples in a bins.
	BinEdgesArray_t 	m_bins;		//!< This is the class that is able to manage the bin edges.
	Numeric_t 		m_sigLevel;	//!< This is the significant level of the that is used, it is impoertant for the bin count
	bool 			m_isMultiVariate; //!< This bool indicates if it is used for multivariate purpose (independence test) or not (GOF).
}; //End class: yggdrasil_chi2TestTable_t





YGGDRASIL_NS_END(yggdrasil)
